# -*- coding: utf-8 -*-
"""
    stego_lsb.StegDetect
    ~~~~~~~~~~~~~~~~~~~~

    This module contains functions for detecting images
    which have been modified using the functions from
    the module :mod:`stego_lsb.LSBSteg`.

    :copyright: (c) 2015 by Ryan Gibson, see AUTHORS.md for more details.
    :license: MIT License, see LICENSE.md for more details.
"""
import logging
import os
from time import time
from typing import cast, Tuple, Iterable

from PIL import Image

log = logging.getLogger(__name__)


def show_lsb(image_path: str, n: int) -> None:
    """Shows the n least significant bits of image"""
    if image_path is None:
        raise ValueError("StegDetect requires an input image file path")

    start = time()
    with Image.open(image_path) as image:
        # Used to set everything but the least significant n bits to 0 when
        # using bitwise AND on an integer
        mask = (1 << n) - 1

        image_data = cast(Iterable[Tuple[int, int, int]], image.getdata())
        color_data = [
            (255 * ((rgb[0] & mask) + (rgb[1] & mask) + (rgb[2] & mask)) // (3 * mask),) * 3
            for rgb in image_data
        ]

        image.putdata(color_data)
        log.debug(f"Runtime: {time() - start:.2f}s")
        file_name, file_extension = os.path.splitext(image_path)
        image.save(f"{file_name}_{n}LSBs{file_extension}")



# -*- coding: utf-8 -*-
"""
    stego_lsb.WavSteg
    ~~~~~~~~~~~~~~~~~

    This module contains functions for hiding and retrieving
    data from .wav files.

    :copyright: (c) 2015 by Ryan Gibson, see AUTHORS.md for more details.
    :license: MIT License, see LICENSE.md for more details.
"""
import logging
import math
import os
import wave
from time import time

from stego_lsb.bit_manipulation import lsb_deinterleave_bytes, lsb_interleave_bytes

log = logging.getLogger(__name__)


def hide_data(sound_path: str, file_path: str, output_path: str, num_lsb: int) -> None:
    """Hide data from the file at file_path in the sound file at sound_path"""
    if sound_path is None:
        raise ValueError("WavSteg hiding requires an input sound file path")
    if file_path is None:
        raise ValueError("WavSteg hiding requires a secret file path")
    if output_path is None:
        raise ValueError("WavSteg hiding requires an output sound file path")

    with wave.open(sound_path, "r") as sound:
        params = sound.getparams()
        num_channels = sound.getnchannels()
        sample_width = sound.getsampwidth()
        num_frames = sound.getnframes()
        num_samples = num_frames * num_channels

        # We can hide up to num_lsb bits in each sample of the sound file
        max_bytes_to_hide = (num_samples * num_lsb) // 8
        file_size = os.stat(file_path).st_size

        log.debug(f"Using {num_lsb} LSBs, we can hide {max_bytes_to_hide} bytes")

        start = time()
        sound_frames = sound.readframes(num_frames)
        with open(file_path, "rb") as file:
            data = file.read()
        log.debug(f"{'Files read':<30} in {time() - start:.2f}s")

        if file_size > max_bytes_to_hide:
            required_lsb = math.ceil(file_size * 8 / num_samples)
            raise ValueError(f"Input file too large to hide, requires {required_lsb} LSBs, using {num_lsb}")

        if sample_width < 1 or sample_width > 4:
            # WavSteg doesn't support higher sample widths, see setsampwidth() in cpython/Libwave.py
            raise ValueError("File has an unsupported bit-depth")

        start = time()
        sound_frames = lsb_interleave_bytes(sound_frames, data, num_lsb, byte_depth=sample_width)
        log.debug(f"{f'{file_size} bytes hidden':<30} in {time() - start:.2f}s")

        start = time()
        with wave.open(output_path, "w") as sound_steg:
            sound_steg.setparams(params)
            sound_steg.writeframes(sound_frames)
        log.debug(f"{'Output wav written':<30} in {time() - start:.2f}s")


def recover_data(sound_path: str, output_path: str, num_lsb: int, bytes_to_recover: int) -> None:
    """Recover data from the file at sound_path to the file at output_path"""
    if sound_path is None:
        raise ValueError("WavSteg recovery requires an input sound file path")
    if output_path is None:
        raise ValueError("WavSteg recovery requires an output file path")
    if bytes_to_recover is None:
        raise ValueError("WavSteg recovery requires the number of bytes to recover")

    start = time()
    with wave.open(sound_path, "r") as sound:
        # num_channels = sound.getnchannels()
        sample_width = sound.getsampwidth()
        num_frames = sound.getnframes()
        sound_frames = sound.readframes(num_frames)
        log.debug(f"{'Files read':<30} in {time() - start:.2f}s")

        if sample_width < 1 or sample_width > 4:
            # WavSteg doesn't support higher sample widths, see setsampwidth() in cpython/Libwave.py
            raise ValueError("File has an unsupported bit-depth")

        start = time()
        data = lsb_deinterleave_bytes(sound_frames, 8 * bytes_to_recover, num_lsb, byte_depth=sample_width)
        log.debug(f"{f'Recovered {bytes_to_recover} bytes':<30} in {time() - start:.2f}s")

        start = time()
        with open(output_path, "wb+") as output_file:
            output_file.write(bytes(data))
        log.debug(f"{'Written output file':<30} in {time() - start:.2f}s")



# -*- coding: utf-8 -*-
"""
stego_lsb.cli
~~~~~~~~~~~~~

This module provides the command line interface for:
    - hiding and recovering data in .wav files
    - hiding and recovering data in bitmap (.bmp and .png)
      files
    - detecting images which have modified using the
      LSB methods.

TODO: should this be refactored more? I am trusting that @sh4nks implemented this well.

:copyright: (c) 2015 by Ryan Gibson, see AUTHORS.md for more details.
:license: MIT License, see LICENSE.md for more details.
"""
import logging

import click

from stego_lsb import LSBSteg, StegDetect, WavSteg, bit_manipulation

# enable logging output
logging.basicConfig(format="%(message)s", level=logging.INFO)
log = logging.getLogger("stego_lsb")
log.setLevel(logging.DEBUG)


@click.group()
@click.version_option()
def main() -> None:
    """Console script for stegolsb."""


@main.command(context_settings=dict(max_content_width=120))
@click.option("--hide", "-h", is_flag=True, help="To hide data in an image file")
@click.option("--recover", "-r", is_flag=True, help="To recover data from an image file")
@click.option("--analyze", "-a", is_flag=True, default=False, show_default=True,
              help="Print how much data can be hidden within an image")
@click.option("--input", "-i", "input_fp", help="Path to an bitmap (.bmp or .png) image")
@click.option("--secret", "-s", "secret_fp", help="Path to a file to hide in the image")
@click.option("--output", "-o", "output_fp", help="Path to an output file")
@click.option("--lsb-count", "-n", default=2, show_default=True, help="How many LSBs to use", type=int)
@click.option("--compression", "-c", help="1 (best speed) to 9 (smallest file size)", default=1, show_default=True,
              type=click.IntRange(1, 9))
@click.pass_context
def steglsb(ctx: click.Context, hide: bool, recover: bool, analyze: bool, input_fp: str, secret_fp: str, output_fp: str,
            lsb_count: int, compression: int) -> None:
    """Hides or recovers data in and from an image"""
    try:
        if analyze:
            LSBSteg.analysis(input_fp, secret_fp, lsb_count)

        if hide:
            LSBSteg.hide_data(input_fp, secret_fp, output_fp, lsb_count, compression)
        elif recover:
            LSBSteg.recover_data(input_fp, output_fp, lsb_count)

        if not hide and not recover and not analyze:
            click.echo(ctx.get_help())
    except ValueError as e:
        log.debug(e)
        click.echo(ctx.get_help())


@main.command()
@click.option("--input", "-i", "image_path", help="Path to an image")
@click.option("--lsb-count", "-n", default=2, show_default=2, type=int, help="How many LSBs to display")
@click.pass_context
def stegdetect(ctx: click.Context, image_path: str, lsb_count: int) -> None:
    """Shows the n least significant bits of image"""
    if image_path:
        StegDetect.show_lsb(image_path, lsb_count)
    else:
        click.echo(ctx.get_help())


@main.command()
@click.option("--hide", "-h", is_flag=True, help="To hide data in a sound file")
@click.option("--recover", "-r", is_flag=True, help="To recover data from a sound file")
@click.option("--input", "-i", "input_fp", help="Path to a .wav file")
@click.option("--secret", "-s", "secret_fp", help="Path to a file to hide in the sound file")
@click.option("--output", "-o", "output_fp", help="Path to an output file")
@click.option("--lsb-count", "-n", default=2, show_default=True, help="How many LSBs to use", type=int)
@click.option("--bytes", "-b", "num_bytes", help="How many bytes to recover from the sound file", type=int)
@click.pass_context
def wavsteg(ctx: click.Context, hide: bool, recover: bool, input_fp: str, secret_fp: str, output_fp: str,
            lsb_count: int, num_bytes: int) -> None:
    """Hides or recovers data in and from a sound file"""
    try:
        if hide:
            WavSteg.hide_data(input_fp, secret_fp, output_fp, lsb_count)
        elif recover:
            WavSteg.recover_data(input_fp, output_fp, lsb_count, num_bytes)
        else:
            click.echo(ctx.get_help())
    except ValueError as e:
        log.debug(e)
        click.echo(ctx.get_help())


@main.command()
def test() -> None:
    """Runs a performance test and verifies decoding consistency"""
    bit_manipulation.test()



import os
import string
import unittest
from random import choice

import numpy as np
import pytest
from PIL import Image

from stego_lsb.LSBSteg import hide_data, recover_data
from stego_lsb.bit_manipulation import roundup


class TestLSBSteg(unittest.TestCase):
    def write_random_image(self, filename: str, width: int, height: int, num_channels: int) -> None:
        image_data = np.random.randint(0, 256, size=(height, width, num_channels), dtype=np.uint8)
        with Image.fromarray(image_data) as image:
            image.save(filename)

    def write_random_file(self, filename: str, num_bytes: int) -> None:
        with open(filename, "wb") as file:
            file.write(os.urandom(num_bytes))

    def check_random_interleaving(self, num_trials: int = 256, filename_length: int = 5, num_channels: int = 3,
                                  skip_storage_check: bool = False, payload_size_shift: int = 0) -> None:
        filename = "".join(choice(string.ascii_lowercase) for _ in range(filename_length))
        png_input_filename = f"{filename}.png"
        payload_filename = f"{filename}.txt"
        png_output_filename = f"{filename}_steg.png"
        recovered_data_filename = f"{filename}_recovered.txt"

        np.random.seed(0)
        for _ in range(num_trials):
            width = np.random.randint(1, 256)
            height = np.random.randint(1, 256)
            num_lsb = np.random.randint(1, 9)

            file_size_tag_length = roundup(int(num_channels * width * height * num_lsb).bit_length() / 8)
            payload_len = (num_channels * width * height * num_lsb - 8 * file_size_tag_length) // 8

            if payload_len < 0:
                continue

            self.write_random_image(png_input_filename, width=width, height=height, num_channels=num_channels)
            self.write_random_file(payload_filename, num_bytes=payload_len + payload_size_shift)

            try:
                hide_data(png_input_filename, payload_filename, png_output_filename, num_lsb, compression_level=1,
                          skip_storage_check=skip_storage_check)
                recover_data(png_output_filename, recovered_data_filename, num_lsb)

                with open(payload_filename, "rb") as input_file, open(recovered_data_filename, "rb") as output_file:
                    input_payload_data = input_file.read()
                    output_payload_data = output_file.read()
            except ValueError as e:
                raise e
            finally:
                for fn in [png_input_filename, payload_filename, png_output_filename, recovered_data_filename]:
                    if os.path.exists(fn):
                        os.remove(fn)

            self.assertEqual(input_payload_data, output_payload_data)

    def test_rgb_steganography_consistency(self) -> None:
        self.check_random_interleaving(num_channels=3)

    def test_rgba_steganography_consistency(self) -> None:
        self.check_random_interleaving(num_channels=4)

    def test_la_steganography_consistency(self) -> None:
        self.check_random_interleaving(num_channels=2)

    def check_maximum_storage(self, num_channels: int = 3) -> None:
        with pytest.raises(ValueError):
            # add an extra byte onto the payload and expect failure
            self.check_random_interleaving(num_channels=num_channels, skip_storage_check=True, payload_size_shift=1)

    def test_rgb_maximum_storage(self) -> None:
        self.check_maximum_storage(num_channels=3)

    def test_rgba_maximum_storage(self) -> None:
        self.check_maximum_storage(num_channels=4)

    def test_la_maximum_storage(self) -> None:
        self.check_maximum_storage(num_channels=2)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
    stego_lsb.LSBSteg
    ~~~~~~~~~~~~~~~~~

    This module contains functions for hiding and recovering
    data from bitmap (.bmp and .png) files.

    :copyright: (c) 2015 by Ryan Gibson, see AUTHORS.md for more details.
    :license: MIT License, see LICENSE.md for more details.
"""
import logging
import os
import sys
from time import time
from typing import Tuple, IO, Union, List, cast

from PIL import Image

from stego_lsb.bit_manipulation import (
    lsb_deinterleave_list,
    lsb_interleave_list,
    roundup,
)

log = logging.getLogger(__name__)


def _str_to_bytes(x: Union[bytes, str], charset: str = sys.getdefaultencoding(), errors: str = "strict") -> bytes:
    if x is None:
        return None
    if isinstance(x, (bytes, bytearray, memoryview)):  # noqa
        return bytes(x)
    if isinstance(x, str):
        return x.encode(charset, errors)
    if isinstance(x, int):
        return str(x).encode(charset, errors)
    raise TypeError("Expected bytes")


def prepare_hide(input_image_path: str, input_file_path: str) -> Tuple[Image.Image, IO[bytes]]:
    """Prepare files for reading and writing for hiding data."""
    # note that these should be closed! consider using context managers instead
    image = Image.open(input_image_path)
    input_file = open(input_file_path, "rb")
    return image, input_file  # these should be closed after use! Consider using a context manager


def prepare_recover(steg_image_path: str, output_file_path: str) -> Tuple[Image.Image, IO[bytes]]:
    """Prepare files for reading and writing for recovering data."""
    # note that these should be closed! consider using context managers instead
    steg_image = Image.open(steg_image_path)
    output_file = open(output_file_path, "wb+")
    return steg_image, output_file  # these should be closed after use! Consider using a context manager


def get_filesize(path: str) -> int:
    """Returns the file size in bytes of the file at path"""
    return os.stat(path).st_size


def max_bits_to_hide(image: Image.Image, num_lsb: int, num_channels: int) -> int:
    """Returns the number of bits we're able to hide in the image using num_lsb least significant bits."""
    # num_channels color channels per pixel, num_lsb bits per color channel.
    return int(num_channels * image.size[0] * image.size[1] * num_lsb)


def bytes_in_max_file_size(image: Image.Image, num_lsb: int, num_channels: int) -> int:
    """Returns the number of bits needed to store the size of the file."""
    return roundup(max_bits_to_hide(image, num_lsb, num_channels).bit_length() / 8)


def hide_message_in_image(input_image: Image.Image, message: Union[str, bytes], num_lsb: int,
                          skip_storage_check: bool = False) -> Image.Image:
    """Hides the message in the input image and returns the modified image object."""
    start = time()
    num_channels = len(input_image.getbands())
    flattened_color_data = [v for t in input_image.getdata() for v in t]

    # We add the size of the input file to the beginning of the payload.
    message_size = len(message)
    file_size_tag = message_size.to_bytes(bytes_in_max_file_size(input_image, num_lsb, num_channels),
                                          byteorder=sys.byteorder)
    data = file_size_tag + _str_to_bytes(message)
    log.debug(f"{'Files read':<30} in {time() - start:.2f}s")

    if 8 * len(data) > max_bits_to_hide(input_image, num_lsb, num_channels) and not skip_storage_check:
        raise ValueError(f"Only able to hide {max_bits_to_hide(input_image, num_lsb, num_channels) // 8} bytes in "
                         f"this image with {num_lsb} LSBs, but {len(data)} bytes were requested")

    start = time()
    flattened_color_data = lsb_interleave_list(flattened_color_data, data, num_lsb)
    log.debug(f"{f'{message_size} bytes hidden':<30} in {time() - start:.2f}s")

    start = time()
    # PIL expects a sequence of tuples, one per pixel
    input_image.putdata(cast(List[int], list(zip(*[iter(flattened_color_data)] * num_channels))))
    log.debug(f"{'Image overwritten':<30} in {time() - start:.2f}s")
    return input_image


def hide_data(input_image_path: str, input_file_path: str, steg_image_path: str, num_lsb: int,
              compression_level: int, skip_storage_check: bool = False) -> None:
    """Hides the data from the input file in the input image."""
    if input_image_path is None:
        raise ValueError("LSBSteg hiding requires an input image file path")
    if input_file_path is None:
        raise ValueError("LSBSteg hiding requires a secret file path")
    if steg_image_path is None:
        raise ValueError("LSBSteg hiding requires an output image file path")

    image, input_file = prepare_hide(input_image_path, input_file_path)
    with image as image, input_file as input_file:
        image = hide_message_in_image(image, input_file.read(), num_lsb, skip_storage_check=skip_storage_check)

        # just in case is_animated is not defined, as suggested by the Pillow documentation
        is_animated = getattr(image, "is_animated", False)
        image.save(steg_image_path, compress_level=compression_level, save_all=is_animated)


def recover_message_from_image(input_image: Image.Image, num_lsb: int) -> bytes:
    """Returns the message from the steganographed image"""
    start = time()
    num_channels = len(input_image.getbands())
    color_data = [v for t in input_image.getdata() for v in t]

    file_size_tag_size = bytes_in_max_file_size(input_image, num_lsb, num_channels)
    tag_bit_height = roundup(8 * file_size_tag_size / num_lsb)

    bytes_to_recover = int.from_bytes(lsb_deinterleave_list(color_data[:tag_bit_height], 8 * file_size_tag_size,
                                                            num_lsb), byteorder=sys.byteorder)

    maximum_bytes_in_image = (max_bits_to_hide(input_image, num_lsb, num_channels) // 8 - file_size_tag_size)
    if bytes_to_recover > maximum_bytes_in_image:
        raise ValueError(f"This image appears to be corrupted.\nIt claims to hold {bytes_to_recover} B, "
                         f"but can only hold {maximum_bytes_in_image} B with {num_lsb} LSBs")

    log.debug(f"{'Files read':<30} in {time() - start:.2f}s")

    start = time()
    data = lsb_deinterleave_list(color_data, 8 * (bytes_to_recover + file_size_tag_size), num_lsb)[
           file_size_tag_size:]
    log.debug(f"{f'{bytes_to_recover} bytes recovered':<30} in {time() - start:.2f}s")
    return data


def recover_data(steg_image_path: str, output_file_path: str, num_lsb: int) -> None:
    """Writes the data from the steganographed image to the output file"""
    if steg_image_path is None:
        raise ValueError("LSBSteg recovery requires an input image file path")
    if output_file_path is None:
        raise ValueError("LSBSteg recovery requires an output file path")

    steg_image, output_file = prepare_recover(steg_image_path, output_file_path)
    with steg_image as steg_image, output_file as output_file:
        data = recover_message_from_image(steg_image, num_lsb)
        start = time()
        output_file.write(data)
        log.debug(f"{'Output file written':<30} in {time() - start:.2f}s")


def analysis(image_file_path: str, input_file_path: str, num_lsb: int) -> None:
    """Print how much data we can hide and the size of the data to be hidden"""
    if image_file_path is None:
        raise ValueError("LSBSteg analysis requires an input image file path")

    with Image.open(image_file_path) as image:
        num_channels = len(image.getbands())
        print(f"Image resolution: ({image.size[0]}, {image.size[1]}, {len(image.getbands())})\n"
              f"{f'Using {num_lsb} LSBs, we can hide:':<30} {max_bits_to_hide(image, num_lsb, num_channels) // 8} B")

        if input_file_path is not None:
            print(f"{'Size of input file:':<30} {get_filesize(input_file_path)} B")

        print(f"{'File size tag:':<30} {bytes_in_max_file_size(image, num_lsb, num_channels)} B")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = [
    "Click", "Pillow", "numpy; python_version>='3.10'",
    "numpy<2.1.0; python_version>='3.9' and python_version<'3.10'",
    "numpy<1.25.0; python_version>='3.8' and python_version<'3.9'"
]

setup(
    author="Ryan Gibson",
    author_email="ryan.alex.gibson@gmail.com",
    name="stego-lsb",
    version="1.6.3",
    description="Least Significant Bit Steganography for bitmap images (.bmp "
                "and .png), WAV sound files, and byte sequences. Simple LSB "
                "Steganalysis (LSB extraction) for bitmap images.",
    keywords="steganography steganalysis",
    license="MIT",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/ragibson/Steganography",
    install_requires=requirements,
    entry_points="""
        [console_scripts]
        stegolsb=stego_lsb.cli:main
    """,
    package_data={"stego_lsb": ["py.typed"]},
    include_package_data=True,
    packages=find_packages(include=["stego_lsb"]),
    zip_safe=False,
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3 :: Only",
    ],
)



import os
import string
import unittest
import wave
from random import choice
from typing import Any, Type

import numpy as np

from stego_lsb.WavSteg import hide_data, recover_data


class TestWavSteg(unittest.TestCase):
    def write_random_wav(self, filename: str, num_channels: int, sample_width: int, framerate: int,
                         num_frames: int) -> None:
        if sample_width < 1 or sample_width > 4:
            # WavSteg doesn't support higher sample widths, see setsampwidth() in cpython/Libwave.py
            raise ValueError("File has an unsupported bit-depth")

        with wave.open(filename, "w") as file:
            file.setnchannels(num_channels)
            file.setsampwidth(sample_width)
            file.setframerate(framerate)

            dtype: Type[np.unsignedinteger[Any]]
            if sample_width == 1:
                dtype = np.uint8
            elif sample_width == 2:
                dtype = np.uint16
            else:
                dtype = np.uint32

            data = np.random.randint(0, 2 ** (8 * sample_width), dtype=dtype, size=num_frames * num_channels)
            # note: typing does not recognize that "writeframes() accepts any bytes-like object" (see documentation)
            file.writeframes(data)  # type: ignore[arg-type,unused-ignore]

    def write_random_file(self, filename: str, num_bytes: int) -> None:
        with open(filename, "wb") as file:
            file.write(os.urandom(num_bytes))

    def check_random_interleaving(self, byte_depth: int = 1, num_trials: int = 256, filename_length: int = 5) -> None:
        filename = "".join(choice(string.ascii_lowercase) for _ in range(filename_length))
        wav_input_filename = f"{filename}.wav"
        payload_input_filename = f"{filename}.txt"
        wav_output_filename = f"{filename}_steg.wav"
        payload_output_filename = f"{filename}_recovered.txt"

        np.random.seed(0)
        for _ in range(num_trials):
            num_channels = np.random.randint(1, 64)
            num_frames = np.random.randint(1, 16384)
            num_lsb = np.random.randint(1, 8 * byte_depth + 1)
            payload_len = (num_frames * num_lsb * num_channels) // 8

            self.write_random_wav(wav_input_filename, num_channels=num_channels, sample_width=byte_depth,
                                  framerate=44100, num_frames=num_frames)
            self.write_random_file(payload_input_filename, num_bytes=payload_len)

            try:
                hide_data(wav_input_filename, payload_input_filename, wav_output_filename, num_lsb)
                recover_data(wav_output_filename, payload_output_filename, num_lsb, payload_len)

                with open(payload_input_filename, "rb") as input_file, open(payload_output_filename,
                                                                            "rb") as output_file:
                    input_payload_data = input_file.read()
                    output_payload_data = output_file.read()
            except ValueError as e:
                raise e
            finally:
                for fn in [wav_input_filename, payload_input_filename, wav_output_filename, payload_output_filename]:
                    if os.path.exists(fn):
                        os.remove(fn)

            self.assertEqual(input_payload_data, output_payload_data)

    def test_consistency_8bit(self) -> None:
        self.check_random_interleaving(byte_depth=1)

    def test_consistency_16bit(self) -> None:
        self.check_random_interleaving(byte_depth=2)

    def test_consistency_24bit(self) -> None:
        self.check_random_interleaving(byte_depth=3)

    def test_consistency_32bit(self) -> None:
        self.check_random_interleaving(byte_depth=4)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np

from stego_lsb.bit_manipulation import lsb_interleave_bytes, lsb_deinterleave_bytes


class TestBitManipulation(unittest.TestCase):
    def assertConsistentInterleaving(self, carrier: bytes, payload: bytes, num_lsb: int, byte_depth: int = 1) -> None:
        num_payload_bits = 8 * len(payload)

        encoded = lsb_interleave_bytes(carrier, payload, num_lsb, byte_depth=byte_depth)
        decoded = lsb_deinterleave_bytes(encoded, num_payload_bits, num_lsb, byte_depth=byte_depth)
        self.assertEqual(decoded, payload)  # payload correctly decoded
        self.assertEqual(len(encoded), len(carrier))  # message length is unchanged after interleaving

        truncated_encode = lsb_interleave_bytes(carrier, payload, num_lsb, byte_depth=byte_depth, truncate=True)
        truncated_decode = lsb_deinterleave_bytes(truncated_encode, num_payload_bits, num_lsb, byte_depth=byte_depth)
        self.assertEqual(truncated_decode, payload)

    def check_random_interleaving(self, byte_depth: int = 1, num_trials: int = 1024) -> None:
        np.random.seed(0)
        for _ in range(num_trials):
            carrier_len = np.random.randint(1, 16384)

            # round up carrier length to next multiple of byte depth
            carrier_len += (byte_depth - carrier_len) % byte_depth
            assert carrier_len % byte_depth == 0

            num_lsb = np.random.randint(1, 8 * byte_depth + 1)
            payload_len = carrier_len * num_lsb // (8 * byte_depth)
            carrier = np.random.randint(0, 256, size=carrier_len, dtype=np.uint8).tobytes()
            payload = np.random.randint(0, 256, size=payload_len, dtype=np.uint8).tobytes()
            self.assertConsistentInterleaving(carrier, payload, num_lsb, byte_depth=byte_depth)

    def test_interleaving_consistency_8bit(self) -> None:
        self.check_random_interleaving(byte_depth=1)

    def test_interleaving_consistency_16bit(self) -> None:
        self.check_random_interleaving(byte_depth=2)

    def test_interleaving_consistency_24bit(self) -> None:
        self.check_random_interleaving(byte_depth=3)

    def test_interleaving_consistency_32bit(self) -> None:
        self.check_random_interleaving(byte_depth=4)

    def test_interleaving_consistency_40bit(self) -> None:
        self.check_random_interleaving(byte_depth=5)

    def test_interleaving_consistency_48bit(self) -> None:
        self.check_random_interleaving(byte_depth=6)

    def test_interleaving_consistency_56bit(self) -> None:
        self.check_random_interleaving(byte_depth=7)

    def test_interleaving_consistency_64bit(self) -> None:
        self.check_random_interleaving(byte_depth=8)


if __name__ == "__main__":
    unittest.main()
